"""LastPass Session class."""
from __future__ import print_function, unicode_literals

import binascii
import hashlib
import json
import logging
import re
import sys
from xml.etree import ElementTree

import requests
import six

from lastpass_aws_login.exceptions import (LastPassCredentialsError, LastPassUnknownError,
                               LastPassIncorrectYubikeyPasswordError,
                               LastPassIncorrectGoogleAuthenticatorCodeError)

LOGGER = logging.getLogger(__name__)


class LastPass(object):
    """LastPass Session management class."""

    def __init__(self, connection_url='https://lastpass.com'):
        self.lastpass_url = 'https://lastpass.com'
        self.connection_url = connection_url

        self.__iterations = 5000
        self.__iterations_username = None
        self.__session = requests.Session()

    def __should_verify(self):
        """Disable SSL validation on connections not addressed to lastpass.com.

        This is to allow connecting use of a proxy server when connecting to
        LastPass
        """
        return self.lastpass_url in self.connection_url

    def __session_get(self, url):
        """Send GET request through internal requests session."""
        LOGGER.debug('[session_get] Sending GET request to %s', url)
        return self.__session.get(url, verify=self.__should_verify())

    def __session_post(self, url, data=None):
        """Send POST request through internal requests session."""
        LOGGER.debug('[session_post] Sending POST request to %s', url)
        return self.__session.post(
            url, data=data, verify=self.__should_verify())

    @staticmethod
    def __extract_form(html):
        """Attempt to extract the first form elements from an html page."""
        fields = {}
        matches = re.findall(r'name="([^"]*)" (id="([^"]*)" )?value="([^"]*)"',
                             html)

        for match in matches:
            if len(match) > 2:
                fields[match[0]] = match[3]

        match = re.search(r'action="([^"]*)"', html)
        action = match.group(1) if match else ''

        return {
            'action': action,
            'fields': fields
        }

    def __get_iterations(self, username):
        """Determine the number of PBKDF2 iterations needed for user."""
        if self.__iterations_username == username:
            LOGGER.debug('[get_iterations] Responding with stored iterations '
                         'for user %s', username)
            return self.__iterations

        LOGGER.debug('[get_iterations] Retrieving iterations for user %s',
                     username)
        iterations_url = '{url}/iterations.php'.format(url=self.connection_url)

        params = {
            'email': username
        }

        response = self.__session_post(iterations_url, data=params)

        if response.status_code == 200:
            self.__iterations = int(response.text)
            self.__iterations_username = username

        return self.__iterations

    @staticmethod
    def __login_hash(username, password, iterations):
        """Determine the login hash for the user."""
        LOGGER.debug('[login_hash] Computing login hash for %s with %d '
                     'iterations', username, iterations)

        if iterations == 1:
            key = binascii.hexlify(hashlib.sha256(username + password).digest())

            return bytearray(hashlib.sha256(key + password).hexdigest(),
                             'ascii')

        key = hashlib.pbkdf2_hmac('sha256', password, username, iterations, 32)

        return binascii.hexlify(hashlib.pbkdf2_hmac('sha256', key, password, 1,
                                                    32))

    @staticmethod
    def __parse_error(parsed_response):
        """Extract error from parsed LastPass response."""
        if parsed_response.tag != 'response':
            error = None
        else:
            error = parsed_response.find('error')

        if error is None or not error.attrib:
            raise LastPassUnknownError('Unknown schema in response from '
                                       'LastPass')

        exceptions = {
            'unknownemail': LastPassCredentialsError,
            'unknownpassword': LastPassCredentialsError,
            'googleauthrequired': LastPassIncorrectGoogleAuthenticatorCodeError,
            'googleauthfailed': LastPassIncorrectGoogleAuthenticatorCodeError,
            'yubikeyrestricted': LastPassIncorrectYubikeyPasswordError
        }

        cause = error.attrib.get('cause')
        message = error.attrib.get('message')

        if cause:
            return exceptions.get(cause, LastPassUnknownError)(message or cause)

        return LastPassUnknownError(message)

    def login(self, username, password, otp=None, client_id=None):
        """Log into LastPass with username, password, and optional OTP code.

        If the user requires an OTP to login a LastPassIncorrectOtpError will be
        raised.
        """
        LOGGER.debug('[login] Starting lastpass login as %s', username)
        iterations = self.__get_iterations(username)
        login_url = '{url}/login.php'.format(url=self.connection_url)

        params = {
            'method': 'mobile',
            'web': 1,
            'xml': 1,
            'username': username,
            'hash': self.__login_hash(username, password, iterations),
            'iterations': iterations
        }

        if otp:
            params['otp'] = otp

        if client_id:
            params['imei'] = client_id

        response = self.__session_post(login_url, data=params)

        if response.status_code != 200:
            LOGGER.debug('[login] Non 200 response from LastPass login: %d',
                         response.status_code)
            raise LastPassUnknownError('Bad response from LastPass')

        try:
            parsed_response = ElementTree.fromstring(response.text)
        except ElementTree.ParseError:
            LOGGER.debug('[login] Error from ElementTree parsing XML response '
                         'from LastPass')
            parsed_response = None

        if parsed_response is None:
            LOGGER.debug('[login] Failed to parse response from LastPass login:'
                         ' %s', response.text)
            raise LastPassUnknownError('Received invalid response from '
                                       'LastPass')

        session_id = None

        if parsed_response.tag == 'ok':
            session_id = parsed_response.attrib.get('sessionid')

            if isinstance(session_id, str):
                return session_id

        LOGGER.debug('[login] No session_id returned, parsing response for '
                     'error')
        raise self.__parse_error(parsed_response)

    def get_saml_token(self, saml_cfg_id):
        """Log into LastPass and retrieve SAML token for config."""
        LOGGER.debug('[get_saml_token] Starting SAML token retrieval')

        # once logged in, grab the SAML token from the IdP-initiated login
        idp_login_url = '{url}/saml/launch/cfg/{saml_cfg_id}'.format(
            url=self.connection_url, saml_cfg_id=saml_cfg_id)

        response = self.__session_get(idp_login_url)
        form = self.__extract_form(response.text)

        if not form['action']:
            # try to scrape the error message to give response to user
            error = ''

            for line in response.text.splitlines():
                match = re.search(r'<h2>(.*)</h2>', line)

                if match:
                    msg = six.moves.html_parser.HTMLParser()\
                        .unescape(match.group(1))
                    msg = msg.replace('<br/>', '\n')
                    msg = msg.replace('<b>', '')
                    msg = msg.replace('</b>', '')
                    error = '\n' + msg

            sys.exit('Unable to find SAML ACS ' + error)

        if not form['fields'].get('SAMLResponse'):
            LOGGER.debug('[get_saml_token] Form: %s', json.dumps(form))
            sys.exit('No SAML response from LastPass')

        return form['fields']['SAMLResponse']



from os import getenv
from os.path import expanduser, join
import sys
import argparse
import json
import logging

try:
    # For Python 3.5 and later
    import configparser
except ImportError:
    # Fall back to Python 2
    import ConfigParser as configparser  # noqa: F401


CALLBACK_SERVER_TIMEOUT = 30


def init():
    aws_config = configparser.ConfigParser()
    home = expanduser("~")
    aws_config.read(join(home, ".aws", "config"))
    try:
        my_conf = Config(_parse_args(), aws_config)
    except ConfigExeption as ex:
        print(ex)
        sys.exit(1)

    return my_conf


def _fail_message(missing_config, profile):
    return "No value for required configuration parameter {} in profile {}.".format(
        missing_config, profile
    )


def _parse_args():
    parser = argparse.ArgumentParser(description="AWS LastPass Login")
    parser.add_argument(
        "-p",
        "--profile",
        type=str,
        nargs="?",
        default=getenv("AWS_PROFILE", "default"),
        help="AWS profile name to log in with",
    )
    parser.add_argument(
        "-u",
        "--user",
        type=str,
        nargs="?",
        default=getenv("LASTPASS_USER", ""),
        help="Username to log in with",
    )
    parser.add_argument(
        "-i",
        "--saml-id",
        type=str,
        nargs="?",
        default=getenv("LASTPASS_SAML_ID", ""),
        help="The lastpass id for the SAML configuration",
    )
    parser.add_argument(
        "-n",
        "--no-prompt",
        action="store_true",
        help="Do not prompt for username and AWS role (need to be set in config in this case)",
    )
    parser.add_argument(
        "-d",
        "--duration",
        type=int,
        help="Duration of the session in hours. Defaults to 1.",
    )
    parser.add_argument("-r", "--role", help="The AWS IAM role ARN to assume")
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose logging")
    return parser.parse_args()


class ConfigExeption(Exception):
    pass


class Config:
    def __init__(self, args, aws_config):
        self.VERBOSE = args.verbose
        self.PROFILE = args.profile
        self.DEFAULT_USERNAME = args.user
        self.CONFIG_PROFILE = args.profile
        self.NO_PROMPT = args.no_prompt
        self.ROLE_ARN = args.role
        self.SAML_ID = args.saml_id
        if args.duration:
            self.DURATION = args.duration * 3600
        else:
            self.DURATION = None
        if self.CONFIG_PROFILE != "default":
            self.CONFIG_PROFILE = "profile {}".format(self.PROFILE)
        if not aws_config.has_section(self.CONFIG_PROFILE):
            raise ConfigExeption(
                "Couldn't find configuration for profile: {}".format(self.PROFILE)
            )

        default_section = config_section = aws_config[self.CONFIG_PROFILE]
        if aws_config.has_section("default"):
            default_section = aws_config["default"]
        if not self.ROLE_ARN:
            self.ROLE_ARN = config_section.get("lastpass_role_arn", default_section.get("lastpass_role_arn", ""))
        if not self.DEFAULT_USERNAME:
            self.DEFAULT_USERNAME = config_section.get("lastpass_default_username", default_section.get("lastpass_default_username", ""))
        if not self.DURATION:
            conf_duration = config_section.get("lastpass_session_duration", default_section.get("lastpass_session_duration", ""))
            if conf_duration.isdigit():
                self.DURATION = int(conf_duration) * 3600
        if not self.DURATION:
            self.DURATION = 3600
        if not self.SAML_ID:
            self.SAML_ID = config_section.get("lastpass_saml_id", default_section.get("lastpass_saml_id", ""))
            if not self.SAML_ID:
                raise ConfigExeption(
                    "Couldn't find SAML configuration id for profile: {}".format(self.PROFILE)
                )
            
    def __str__(self):
        return json.dumps(
            {
                "profile": self.PROFILE,
                "username": self.DEFAULT_USERNAME,
                "profile": self.PROFILE,
                "conf_profile": self.CONFIG_PROFILE,
                "no-prompt": self.NO_PROMPT,
                "role-arn": self.ROLE_ARN,
                "saml-id": self.SAML_ID,
            },
            indent=2,
        )


if __name__ == "__main__":
    init()



"""Custom exception classes for project."""


# LastPass errors
class LastPassError(Exception):
    """General error for LastPass class."""


class LastPassCredentialsError(LastPassError):
    """Error for invalid credentials provided."""


class LastPassUnknownError(LastPassError):
    """Unknown error when handling LastPass connection."""


class LastPassIncorrectOtpError(LastPassError):
    """General error for incorrect OTP codes."""


class LastPassIncorrectGoogleAuthenticatorCodeError(LastPassIncorrectOtpError):
    """LastPass error for missing or incorrect Google Authenticator code."""


class LastPassIncorrectYubikeyPasswordError(LastPassIncorrectOtpError):
    """LastPass error for missing or incorrect Yubikey password."""



"""Static utility functions."""
import sys
import xml.etree.ElementTree as ET


def binary_type(string):
    """Return binary_type of string."""
    if sys.version_info[0] == 2:
        return string

    return string.encode('utf-8')


def get_saml_aws_roles(assertion):
    """Get the AWS roles contained in a decoded SAML assertion.

    This returns a list of RoleARN, PrincipalARN (IdP) pairs.
    """
    # Parse the returned assertion and extract the authorized roles
    awsroles = []
    root = ET.fromstring(assertion)
    for saml2attribute in root.iter("{urn:oasis:names:tc:SAML:2.0:assertion}Attribute"):
        if saml2attribute.get("Name") == "https://aws.amazon.com/SAML/Attributes/Role":
            for saml2attributevalue in saml2attribute.iter(
                "{urn:oasis:names:tc:SAML:2.0:assertion}AttributeValue"
            ):
                awsroles.append(saml2attributevalue.text)

    for awsrole in awsroles:
        chunks = awsrole.split(",")
        if "saml-provider" in chunks[0]:
            newawsrole = chunks[1] + "," + chunks[0]
            index = awsroles.index(awsrole)
            awsroles.insert(index, newawsrole)
            awsroles.remove(awsrole)

    return awsroles


from os.path import expanduser

try:
    # For Python 3.5 and later
    import configparser
except ImportError:
    # Fall back to Python 2
    import ConfigParser as configparser


def write(new_creds, profile):
    aws_creds = configparser.ConfigParser()
    home = expanduser("~")
    creds_file = home + "/.aws/credentials"
    aws_creds.read(creds_file)

    if not aws_creds.has_section(profile):
        aws_creds.add_section(profile)

    c = new_creds["Credentials"]
    aws_creds.set(profile, "aws_access_key_id", c["AccessKeyId"])
    aws_creds.set(profile, "aws_secret_access_key", c["SecretAccessKey"])
    aws_creds.set(profile, "aws_session_token", c["SessionToken"])
    aws_creds.set(profile, "aws_session_expiration", c["Expiration"].isoformat())

    with open(creds_file, "w") as configfile:
        aws_creds.write(configfile)



"""aws-lp main"""
from __future__ import print_function, unicode_literals

import base64
import logging
import sys
from os import environ
from builtins import input
from getpass import getpass

from lastpass_aws_login import credentials
from lastpass_aws_login.conf import init
from lastpass_aws_login.exceptions import (LastPassCredentialsError, LastPassError,
                               LastPassIncorrectOtpError)
from lastpass_aws_login.lastpass import LastPass
from lastpass_aws_login.utils import binary_type, get_saml_aws_roles

from threadlocal_aws.clients import sts

logging.basicConfig(
    format='[%(asctime)s][%(name)s][%(levelname)s]: %(message)s',
    stream=sys.stdout)
LOGGER = logging.getLogger(__name__)


def main():
    """Log into LastPass, get SAML auth, assume role, and write tokens into credentials file"""
    conf = init()

    if conf.VERBOSE:
        logging.getLogger('lastpass_aws_login').setLevel(logging.DEBUG)


    username = None
    # Get the federated credentials from the user
    if not conf.NO_PROMPT:
        sys.stdout.write("Username [" + conf.DEFAULT_USERNAME + "]: ")
        username = input()
    if not username:
        if conf.DEFAULT_USERNAME:
            username = conf.DEFAULT_USERNAME
        else:
            print("Need to give username")
            sys.exit(1)
    if "LASTPASS_DEFAULT_PASSWORD" in environ and environ["LASTPASS_DEFAULT_PASSWORD"]:
        password = environ["LASTPASS_DEFAULT_PASSWORD"]
    else:
        password = getpass()

    username = binary_type(username)
    password = binary_type(password)
    lastpass_session = LastPass('https://lastpass.com')

    try:
        lastpass_session.login(username, password)
    except LastPassIncorrectOtpError:
        mfa = input('MFA: ')

        try:
            lastpass_session.login(username, password, otp=mfa)
        except LastPassIncorrectOtpError:
            sys.exit('Invalid MFA code')
    except LastPassCredentialsError:
        sys.exit('Invalid username or password')
    except LastPassError as error:
        # don't display stack trace but still exit and print error message
        sys.exit(str(error))

    assertion = lastpass_session.get_saml_token(conf.SAML_ID)

    awsroles = get_saml_aws_roles(base64.b64decode(assertion))
    # Overwrite and delete the credential variables, just for safety
    username = "##############################################"
    password = "##############################################"
    del username
    del password
    role_arn = None
    if conf.NO_PROMPT and conf.ROLE_ARN:
        for awsrole in awsroles:
            if awsrole.startswith(conf.ROLE_ARN + ","):
                role_arn = conf.ROLE_ARN
                principal_arn = awsrole.split(",")[1]
        if not role_arn:
           role_arn, principal_arn = select_role(awsroles)
    else:
        # If I have more than one role, ask the user which one they want,
        # otherwise just proceed
       role_arn, principal_arn = select_role(awsroles)

    if not role_arn:
        print("No valid role found in assertions")
        print(awsroles)
        sys.exit(3)
    # Use the assertion to get an AWS STS token using Assume Role with SAML
    token = sts().assume_role_with_saml(
        RoleArn=role_arn,
        PrincipalArn=principal_arn,
        SAMLAssertion=assertion,
        DurationSeconds=conf.DURATION,
    )
    credentials.write(token, conf.PROFILE)

def select_role(awsroles):
    role_arn = None
    principal_arn = None
    if len(awsroles) > 1:
        i = 0
        print("Please choose the role you would like to assume:")
        for awsrole in awsroles:
            print("[" + str(i) + "]: " + awsrole.split(",")[0])
            i += 1
        sys.stdout.write("Selection: ")
        selectedroleindex = input()

        # Basic sanity check of input
        if int(selectedroleindex) > (len(awsroles) - 1):
            print("You selected an invalid role index, please try again")
            sys.exit(1)

        role_arn = awsroles[int(selectedroleindex)].split(",")[0]
        principal_arn = awsroles[int(selectedroleindex)].split(",")[1]
    elif awsroles:
        role_arn = awsroles[0].split(",")[0]
        principal_arn = awsroles[0].split(",")[1]
    return role_arn, principal_arn



"""setuptools installation script"""
import io
import sys
from os import path

from setuptools import setup, find_packages


with io.open(path.join(path.abspath(path.dirname(__file__)), 'README.md'),
             encoding='utf-8') as readme:
    LONG_DESCRIPTION = readme.read()

if sys.version_info[0] == 2:
    python2_or_3_deps = ['importlib-metadata==2.1.1', 'decorator==4.4.2']
    python2_or_3_test_deps = ['pytest==4.6.11', 'pytest-mock==1.13.0', 'mock==3.0.5']
elif sys.version_info[0] == 3:
    python2_or_3_deps = []
    python2_or_3_test_deps = ['pytest-mock', 'mock']
    if sys.version_info[1] == 5:
        python2_or_3_deps = ["boto3==1.16.63"]
        python2_or_3_test_deps.insert(0, "pytest==6.1.2")
        python2_or_3_test_deps.append('importlib-metadata==2.1.1')
    else:
        python2_or_3_test_deps.insert(0, "pytest")


setup(
    name='lastpass-aws-login',
    version="0.1.2",
    description='Tool for using AWS CLI with LastPass SAML',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author="Pasi Niemi",
    author_email='pasi.niemi@nitor.com',
    url='https://github.com/NitorCreations/lastpass-aws-login',
    license='GPLv3',
    keywords='lastpass aws awscli boto3',
    packages=['lastpass_aws_login'],
    setup_requires=["pytest-runner"],
    install_requires=[
        'threadlocal-aws==0.8',
        'requests>=2.22.0"',
        'future',
    ] + python2_or_3_deps,
    entry_points={
        'console_scripts': [
            'lastpass-aws-login=lastpass_aws_login.main:main'
        ]
    },
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    tests_require=[
        "requests-mock==1.8.0",
        "pytest-cov==2.11.1",
    ] + python2_or_3_test_deps,
    test_suite="tests",
)
